#!/usr/bin/env python3

import sys
import os
import argparse
import subprocess
import gzip


def get_args():
    parser = argparse.ArgumentParser(
        description="Utility script to merge VCFs from MRJD and VC into a single VCF"
    )
    optional = parser._action_groups.pop()
    required = parser.add_argument_group("required arguments")

    # required
    required.add_argument(
        "--bed",
        type=str,
        help="MRJD bed file",
        required=True,
    )

    required.add_argument(
        "--vc",
        type=str,
        help="VCF file path from DRAGEN VC",
        required=True,
    )
    required.add_argument(
        "--mrjd",
        type=str,
        help="VCF file path from DRAGEN MRJD",
        required=True,
    )
    # optional
    optional.add_argument(
        "-o",
        "--outdir",
        type=str,
        help="Output file directory (default = '.')",
        required=False,
    )
    parser._action_groups.append(optional)
    args = parser.parse_args()

    # check if input BED file exists
    args.bed = os.path.abspath(args.bed)
    try:
        print("Input BED file: " + args.bed)
        test = open(args.bed, "r")
    except Exception as e:
        print("Can not open input BED file: " + args.bed)
        sys.exit(1)

    # check if input DRAGEN VC VCF file exists
    args.vc = os.path.abspath(args.vc)
    try:
        print("Input DRAGEN VC VCF file: " + args.vc)
        test = open(args.vc, "r")
    except Exception as e:
        print("Can not open input DRAGEN VC VCF file: " + args.vc)
        sys.exit(1)

    # check if input DRAGEN MRJD VCF file exists
    args.mrjd = os.path.abspath(args.mrjd)
    try:
        print("Input DRAGEN MRJD VCF file: " + args.mrjd)
        test = open(args.mrjd, "r")
    except Exception as e:
        print("Can not open input DRAGEN MRJD VCF file: " + args.mrjd)
        sys.exit(1)

    # sets up default output directory
    if args.outdir is None:
        args.outdir = "."
    args.outdir = os.path.abspath(args.outdir)
    if not os.path.isdir(args.outdir):
        try:
            os.mkdir(args.outdir)
        except OSError:
            print("Creation of the directory %s failed" % args.outdir)
        else:
            print("Successfully created the directory %s " % args.outdir)

    print("\n")

    return args


def add_info_to_vcf(input_vcf_gz, output_vcf_gz):
    with gzip.open(input_vcf_gz, "rt") as f_in, open(output_vcf_gz, "wt") as f_out:
        mrjd_header_added = False
        for line in f_in:
            if line.startswith("##INFO"):
                if not mrjd_header_added:
                    f_out.write(
                        '##INFO=<ID=MRJD,Number=0,Type=Flag,Description="Calls made by MRJD">\n'
                    )
                    mrjd_header_added = True
            elif line.startswith("#"):
                f_out.write(line)
            else:
                fields = line.strip().split("\t")
                info = fields[7]
                if info == ".":
                    fields[7] = "MRJD"
                else:
                    fields[7] = "MRJD;" + fields[7]
                f_out.write("\t".join(fields) + "\n")


def extract_header(vcf_file):
    """Extract header lines from a VCF file."""
    header_lines = []
    with gzip.open(vcf_file, "rt") as f:
        for line in f:
            if line.startswith("##"):
                header_lines.append(line)
            elif line.startswith("#CHROM"):
                header_lines.append(line)
                break
    return header_lines


def modify_header(header_lines, additional_lines):
    """Modify header by adding additional lines after INFO lines."""
    modified_header = []
    info_lines_added = False
    for line in header_lines:
        if line.startswith("##INFO"):
            modified_header.append(line)
            if not info_lines_added:
                for additional_line in additional_lines:
                    modified_header.append(additional_line + "\n")
                info_lines_added = True
        else:
            modified_header.append(line)
    return modified_header


def write_modified_header(modified_header, output_file):
    """Write modified header to a new file."""
    with open(output_file, "w") as f:
        for line in modified_header:
            f.write(line)


def main():
    args = get_args()

    # Remove VC entries that overlap the MRJD bed
    sample_name = os.path.basename(args.vc).split(".")[0]
    vc_filtered_file_basename = os.path.basename(args.vc).replace(
        ".vcf.gz", ".vc_filtered.vcf.gz"
    )
    vc_filtered_file_path = args.outdir + "/" + vc_filtered_file_basename

    cmd = f"bcftools view -T ^{args.bed} {args.vc} -Oz > {vc_filtered_file_path}"
    subprocess.run(cmd, shell=True, check=True)

    cmd = f"tabix -f {vc_filtered_file_path}"
    subprocess.run(cmd, shell=True, check=True)

    # Modify the MRJD VCF to add MRJD to the INFO column
    mrjd_modified_vcf_basename = os.path.basename(args.mrjd).replace(
        ".vcf.gz", ".mrjd_tag_added.vcf"
    )
    mrjd_modified_vcf_path = args.outdir + "/" + mrjd_modified_vcf_basename
    add_info_to_vcf(args.mrjd, mrjd_modified_vcf_path)
    mrjd_modified_vcf_gz_path = mrjd_modified_vcf_path + ".gz"

    subprocess.run(
        ["bgzip", "-c", mrjd_modified_vcf_path],
        stdout=open(mrjd_modified_vcf_gz_path, "wb"),
    )
    os.remove(mrjd_modified_vcf_path)
    cmd = f"tabix -f {mrjd_modified_vcf_gz_path}"
    subprocess.run(cmd, shell=True, check=True)

    # Because the VC vcf.gz file has a set of headers and the MRJD vcf.gz has a different set of headers. I need to merge those headers.
    new_header = args.outdir + "/" + "modified_header.txt"

    additional_lines = [
        '##FORMAT=<ID=JAD,Number=R,Type=Integer,Description="Allelic depths (counting only informative reads out of the total reads) for the ref and alt alleles in the order listed. The Allelic depth is the joint depth across all paralog copies">',
        '##FORMAT=<ID=JAF,Number=A,Type=Float,Description="Allele fractions for alt alleles in the order listed. The Allelic fraction is considering all reads across all paralog copies">',
        '##FORMAT=<ID=JDP,Number=1,Type=Integer,Description="Approximate read depth across all paralog copies">',
        '##FORMAT=<ID=JQL,Number=1,Type=Float,Description="Joint quality, phred-scaled prob(call in ALT is wrong) in all paralogous regions">',
        '##FORMAT=<ID=JGT,Number=1,Type=String,Description="Joint Genotype in all paralogous regions for ambiguous calls">',
        '##FORMAT=<ID=JGQ,Number=1,Type=Integer,Description="Joint Genotype Quality">',
        '##FORMAT=<ID=JPL,Number=.,Type=Integer,Description="Normalized, Phred-scaled likelihoods for joint genotypes across all paralogous regions">',
        '##INFO=<ID=MRJD,Number=0,Type=Flag,Description="Calls made by MRJD">',
        '##INFO=<ID=JIDS,Number=0,Type=Flag,Description="Joint IDs">',
        '##INFO=<ID=REGION_AMBIGUOUS,Number=0,Type=Flag,Description="Region-ambiguous calls">',
        '##INFO=<ID=MRJD_HS,Number=0,Type=Flag,Description="Calls made by MRJD high sensitivity mode">',
    ]

    header_lines = extract_header(vc_filtered_file_path)
    modified_header = modify_header(header_lines, additional_lines)
    write_modified_header(modified_header, new_header)

    # Add the header to the filtered vcf file
    vc_filtered_reheadered_file_path = vc_filtered_file_path.replace(
        ".vcf.gz", ".reheadered.vcf.gz"
    )

    cmd = f"bcftools reheader --header {new_header} {vc_filtered_file_path} -o {vc_filtered_reheadered_file_path}"
    subprocess.run(cmd, shell=True, check=True)
    cmd = f"tabix -f {vc_filtered_reheadered_file_path}"
    subprocess.run(cmd, shell=True, check=True)

    vc_filtered_reheadered_sorted_file_path = vc_filtered_file_path.replace(
        ".vcf.gz", ".reheadered.sorted.vcf.gz"
    )
    cmd = f"bcftools sort {vc_filtered_reheadered_file_path} -o {vc_filtered_reheadered_sorted_file_path}"
    subprocess.run(cmd, shell=True, check=True)
    cmd = f"tabix -f {vc_filtered_reheadered_sorted_file_path}"
    subprocess.run(cmd, shell=True, check=True)

    # Add the header to the MRJD modified vcf file
    mrjd_modified_reheadered_vcf_gz_path = mrjd_modified_vcf_gz_path.replace(
        ".vcf.gz", ".reheadered.vcf.gz"
    )

    cmd = f"bcftools reheader --header {new_header} {mrjd_modified_vcf_gz_path} -o {mrjd_modified_reheadered_vcf_gz_path}"
    subprocess.run(cmd, shell=True, check=True)
    cmd = f"tabix -f {mrjd_modified_reheadered_vcf_gz_path}"
    subprocess.run(cmd, shell=True, check=True)

    mrjd_modified_reheadered_sorted_vcf_gz_path = mrjd_modified_vcf_gz_path.replace(
        ".vcf.gz", ".reheadered.sorted.vcf.gz"
    )
    cmd = f"bcftools sort {mrjd_modified_reheadered_vcf_gz_path} -o {mrjd_modified_reheadered_sorted_vcf_gz_path}"
    subprocess.run(cmd, shell=True, check=True)
    cmd = f"tabix -f {mrjd_modified_reheadered_sorted_vcf_gz_path}"
    subprocess.run(cmd, shell=True, check=True)

    merged_file_path = args.outdir + "/" + "merged.vcf.gz"
    cmd = f"bcftools concat {vc_filtered_reheadered_sorted_file_path} {mrjd_modified_reheadered_sorted_vcf_gz_path} -o {merged_file_path} -a"
    subprocess.run(cmd, shell=True, check=True)

    cmd = f"tabix -f {merged_file_path}"
    subprocess.run(cmd, shell=True, check=True)

    merged_sorted_file_path = args.outdir + "/" + sample_name + ".vc_mrjd_merged.vcf.gz"
    cmd = f"bcftools sort {merged_file_path} -o {merged_sorted_file_path}"
    subprocess.run(cmd, shell=True, check=True)

    cmd = f"tabix -f {merged_sorted_file_path}"
    subprocess.run(cmd, shell=True, check=True)

    # Clean files
    os.remove(merged_file_path)
    os.remove(merged_file_path + ".tbi")

    os.remove(vc_filtered_reheadered_file_path)
    os.remove(vc_filtered_reheadered_file_path + ".tbi")

    os.remove(mrjd_modified_reheadered_vcf_gz_path)
    os.remove(mrjd_modified_reheadered_vcf_gz_path + ".tbi")

    os.remove(new_header)

    os.remove(vc_filtered_reheadered_sorted_file_path)
    os.remove(vc_filtered_reheadered_sorted_file_path + ".tbi")

    os.remove(mrjd_modified_reheadered_sorted_vcf_gz_path)
    os.remove(mrjd_modified_reheadered_sorted_vcf_gz_path + ".tbi")

    os.remove(mrjd_modified_vcf_gz_path)
    os.remove(mrjd_modified_vcf_gz_path + ".tbi")

    os.remove(vc_filtered_file_path)
    os.remove(vc_filtered_file_path + ".tbi")


if __name__ == "__main__":
    main()
